//*----------------------------------------------------------------------------
//*	ATMEL Microcontroller Software Support - ROUSSET -
//*----------------------------------------------------------------------------
//* The software is delivered "AS IS" without warranty or condition of any
//* kind, either express, implied or statutory. This includes without
//* limitation any warranty or condition with respect to merchantability or
//* fitness for any particular purpose, or against the infringements of
//* intellectual property rights of others.
//*----------------------------------------------------------------------------
//* File Name	: init.c
//* Object	: Low level initialisations written in C
//* Creation	: ODi   06/26/2002
//*
//*----------------------------------------------------------------------------

/* modified 2004-2007 by Dmitriy Cherkashin */
/* email: divch@users.sourceforge.net, dch@ucrouter.ru */
/* site: www.ucrouter.ru phone: +7-910-422-2917 */
/* based on ATMEL AT91RM9200-BasicBoot */
/* and AT91RM9200-BasicROM_Services code */
/* MIT licence */

#include "AT91RM9200.h"

#ifdef __GNUC__
#include "proto.h"			
#else 
#include "lib_AT91RM9200.h"
#endif							/* __GNUC__ */

/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////

#define	FALSE				0
#define	TRUE				1

#define DELAY_ON
#ifdef  DELAY_ON
#define	DELAY_PLL			100
#define DELAY_MAIN_FREQ		100
#endif 

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Configuration for a Quartz 18.432000 MHz
/////////////////////////////////////////////////////////////////////////////////////////////////////

//#define	PLLAR 			0x20263E04	//* 179,712000 MHz for PCK
/*
0-7  - DIVA    
8-13 - PLLACOUNT
15-14 - OUTA - PLL A Clock Frequency Range
	00 - 80-160MHz
	01 - reserved
	10 - 150-240MHz
	11 - reserved

26-16 - MULA

0x2026BE04
div  = 04
PLLACOUNT = 3E
OUTA = 00
MULA = 26 (38) 
*/


#define PLLA_DIV     0x04		/* 18432000 / 4 -> 4608000*/
#define PLLA_COUNT   0x3E
#define PLLA_OUTA    0x02
#define PLLA_MULA    0x26		/* 26 = 38 -> x 39 = 179.712000 */
//#define PLLA_MULA    0x2A		/* 2B = 43 -> x 44 = 202.752000 */
//#define PLLA_MULA    0x2C		/* 2C = 44 -> x 45 = 207.360000 */		

#define PLLAR						\
   (((PLLA_DIV    & 0xFF) <<  0) |	\
	((PLLA_COUNT  & 0x3F) <<  8) |	\
	((PLLA_OUTA   & 0x03) << 14) |	\
	((PLLA_MULA   & 0x7F) << 16) |	\
	 0x20000000)

#define	PLLAR_OLD 		0x2026BE04	//* 179,712000 MHz for PCK

#define PLLB_DIV     0x0E
#define PLLB_COUNT   0x3E
#define PLLB_OUTA    0x00
#define PLLB_MULA    0x48
#define PLLA_USB_96M 0x01 


#define PLLBR						\
   (((PLLB_DIV    & 0xFF) <<  0) |	\
	((PLLB_COUNT  & 0x3F) <<  8) |	\
	((PLLB_OUTA   & 0x03) << 14) |	\
	((PLLB_MULA   & 0x7F) << 16) |	\
	((PLLA_USB_96M& 0x01) << 28) )

#define	PLLBR_OLD 			0x10483E0E	//* 48,054857 MHz (divider by 2 for USB)

/* #define OLD_CLK */ 					// uncomment to work at 60Mhz

#ifdef OLD_CLK
#define	MCKR  			0x00000202	// PCK/3 = MCK Master Clock = 59,904000MHz with PLLA selected
#define SDRC_TR_VAL     (64 * 59904 / 4096)	// (64 / 1000 / 4096 * 59904000)
#else 
/*
 1-0 - CSS Clock source selection
 4-2 - PRES master clock prescaler 
 		0 - selected clock
 		1 - selected clock divided 2
 		2 - divided by 4  
		6 - divided by 64
  9-8 - MDIV - master clock divisor 
  		0 - master clock and processor clock are the same 
		1 - processor clock is twice fast 
		2 - processor clock is 3 times faster then the master clock 
		3 - processor clock is 4 times faster then the master clock 
*/
#define	MCKR_CSS      0x02				// clock source selection 
#define	MCKR_PRES     0x00				// prescaler 
//#define	MCKR_MDIV     0x01				// master clock divisor 
//#define	MCKR_MDIV     0x02				// master clock divisor 
#define	MCKR_MDIV     0x03				// master clock divisor 

#define	MCKR					\
	   (((MCKR_CSS & 0x03) << 0) |		\
		((MCKR_PRES & 0x7) << 2) |	\
		((MCKR_MDIV & 0x3) << 8))

#define	MCKR_OLD			0x00000302	// PCK/4 =
// (64 / 1000 / 4096 * 44928000)
#define SDRC_TR_VAL  	(64 * ((18432 / PLLA_DIV * (PLLA_MULA + 1)) / (MCKR_MDIV + 1)) / 4096)
#endif

#define SLOWCLOCK		32768		//* In Hz

extern unsigned long master_clock;
unsigned long sdram_twr;
unsigned long sdram_trc;
unsigned long sdram_trp;
unsigned long sdram_trcd;
unsigned long sdram_tras;
unsigned long sdram_txsr;
unsigned long sdram_clk;


/////////////////////////////////////////////////////////////////////////////////////////////////////
// Frequencies Range
/////////////////////////////////////////////////////////////////////////////////////////////////////

#define INPUT_FREQ_MIN		900000
#define INPUT_FREQ_MAX		32000000

#ifdef AT91RM9200

#define BASE_EBI_CS0_ADDRESS	0x10000000	//* base address to access memory on CS0
#define BASE_EBI_CS1_ADDRESS	0x20000000	//* base address to access memory on CS1

#define OUTPUT_FREQ_MIN		80000000
#define OUTPUT_FREQ_MAX		240000000

#else
#ifdef AT91RM3400
	
#define OUTPUT_FREQ_MIN		20000000
#define OUTPUT_FREQ_MAX		100000000

#endif
#endif

extern void AT91F_DBGU_Printx(unsigned long val);
extern void AT91F_DBGU_Printl(unsigned long val);

//*----------------------------------------------------------------------------
//* \fn    AT91F_DBGU_Printk
//* \brief This function is used to send a string through the DBGU channel (Very low level debugging)
//*----------------------------------------------------------------------------
void AT91F_DBGU_Printk(char *buffer) // \arg pointer to a string ending by \0
{
 char ch; 							// readed character
 if(buffer == 0) {
	return;
 }

 ch = *buffer++; 					// read character from buffer
 while(ch != '\0') {
	// wait tx ready and print character 
	while (!AT91F_US_TxReady((AT91PS_USART)AT91C_BASE_DBGU));
	AT91F_US_PutChar((AT91PS_USART)AT91C_BASE_DBGU, ch);

	if(ch == '\n') {
		while (!AT91F_US_TxReady((AT91PS_USART)AT91C_BASE_DBGU));
		ch = '\r';
	 	AT91F_US_PutChar((AT91PS_USART)AT91C_BASE_DBGU, ch);
	}
	ch = *buffer++; 				// read character from buffer 
 }
}

/* get character througth DBGU port */

int AT91F_DBGU_getch(char * ch)
{
 if(AT91F_US_RxReady( (AT91PS_USART)AT91C_BASE_DBGU ) == 0) {
    return(0);
 } else {
    if(ch != 0) {
	*ch = AT91F_US_GetChar( (AT91PS_USART)AT91C_BASE_DBGU );
    }
    return(1);
 }
}

/* */

void AT91F_DBGU_Printv(char * pref, unsigned long val, char *suf)
{
 if(pref != 0) {
 	AT91F_DBGU_Printk(pref);
	AT91F_DBGU_Printk(" ");
 }
 AT91F_DBGU_Printl(val);
 if(suf != 0) {
	AT91F_DBGU_Printk(suf);
 } else {
 	AT91F_DBGU_Printk("\n");
 }
}

void AT91F_DBGU_Printw(char * pref, unsigned long val, char *suf)
{
 if(pref != 0) {
 	AT91F_DBGU_Printk(pref);
 	AT91F_DBGU_Printk(" ");
 }
 AT91F_DBGU_Printx(val);
 if(suf != 0) {
	AT91F_DBGU_Printk(suf);
 } else {
 	AT91F_DBGU_Printk("\n");
 }
}

/* print long value */

void AT91F_DBGU_Printl(unsigned long val)
{
 char buf[17];				// temp buffer  
 int pos;					// buffer position 
 int dpos;					// buffer destination position 

 buf[16] = '\0';			// terminating character 
 pos = 15;					// current buffer position 

 while(val > 0 && pos >= 0) {
    char ch;
    ch = (val % 10) + '0'; 	// next character writed to buffer 
    buf[pos] = ch;			// write character to buffer
    val = val / 10;			// next decimal value 	
    pos--;					// next buffer position 
 }
 pos++;									

 dpos = 0;
 while(pos <= 16) {
    buf[dpos++] = buf[pos++];  
 }

 AT91F_DBGU_Printk(buf);
} 

/* print hexadecimal value */

void AT91F_DBGU_Printx(unsigned long val)
{
 char buf[12];
 int pos;
 for(pos = 0; pos < 8; ++pos) {
    int ch;
    ch = (val >> ((7 - pos) * 4)) & 0xF;	
    if(ch >= 0 && ch <= 9) {
    	ch = '0' + ch;
    } else {
		ch = 'A' + (ch - 10);
    }
    buf[pos] = ch;
 }
 buf[8] = '\n';
 buf[9] = '\0';
 AT91F_DBGU_Printk(buf);
} 

/* read unsigned long value from console */

void AT91F_DBGU_getx(unsigned long * val)
{
 char buf[12];					/* buffer */
 int pos;						/* buffer position */
 char it;

 if(val == 0) {
	return;
 }

 pos = 0;
 buf[pos] = '\0';
 while(pos < 8) {
	it = 0;
	if(AT91F_DBGU_getch( & it )) {
		if(it >= 'a' && it <= 'z') {
			it = 'A' + (it - 'a');
	   	}

		if( (it >= '0' && it <= '9') || (it >= 'A' && it <= 'Z') ) {
			buf[pos  ] = it;
			buf[pos+1] = '\0';
			AT91F_DBGU_Printk(buf + pos);
			pos++;
		} else
		if(it == '\b' && pos > 0) {
			AT91F_DBGU_Printk("\b \b");
			pos--;
			buf[pos] = '\0';
		} else 
		if(it == '\n' || it == '\r') {
			break;
		}
	}
 }

 *val = 0;

 for(pos = 0; (pos < 12 && buf[pos] != '\0'); ++pos) {
	it = buf[pos];
	if(it >= '0' && it <= '9') {
		it -= '0'; 
	} else {
		it = 10 + (it - 'A');
	}

	*val = (*val << 4) | it; 
 }
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_DataAbort
//* \brief This function reports an Abort
//*----------------------------------------------------------------------------

void AT91F_SpuriousHandler(void)
{
    AT91F_DBGU_Printk("-O- Spurious Int\n");
    while (1);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_DataAbort
//* \brief This function reports an Abort
//*----------------------------------------------------------------------------

void AT91F_DataAbort(void)
{
    AT91F_DBGU_Printk("-D- Data Abort\n");
    while (1);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_FetchAbort
//* \brief This function reports an Abort
//*----------------------------------------------------------------------------
void AT91F_FetchAbort(void)
{
    AT91F_DBGU_Printk("-P- Prefetch Abort\n");
    while (1);
}

//*----------------------------------------------------------------------------
//* default IRQ handler
//*----------------------------------------------------------------------------
void AT91F_IrqHandler(void)
{
    AT91F_DBGU_Printk("-I- IRQ\n");
    while (1);
}

//*----------------------------------------------------------------------------
//  FIQ handler
//*----------------------------------------------------------------------------
void AT91F_FiqHandler(void)
{
    AT91F_DBGU_Printk("-F- FIQ\n");
    while (1);
}

//*----------------------------------------------------------------------------
//* SWI handler
//*----------------------------------------------------------------------------
void AT91F_SwiHandler(void)
{
    AT91F_DBGU_Printk("-S- \n");
    while (1);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_UndefHandler
//* \brief This function reports that no handler have been set for current IT
//*----------------------------------------------------------------------------
void AT91F_UndefHandler(void)
{
    AT91F_DBGU_Printk("-U- Undef\n");
    while (1);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////

//*----------------------------------------------------------------------------
//* \fn    AT91F_CheckPLL_FrequencyRange
//* \brief This function performs very low level HW initialiszation
//*----------------------------------------------------------------------------
unsigned char AT91F_CheckPLL_FrequencyRange(unsigned int MainClock,unsigned int pllDivider ,unsigned int pllMultiplier)
{
    if(pllDivider == 0) {
	return FALSE;
    }

    //* Check Input Frequency
    if( ((MainClock/pllDivider) < INPUT_FREQ_MIN)
     || ((MainClock/pllDivider) > INPUT_FREQ_MAX) ) {
    	return FALSE;
    }
    
    //* Check Output Frequency
    if( ((MainClock/pllDivider*pllMultiplier) < OUTPUT_FREQ_MIN)
     || ((MainClock/pllDivider*pllMultiplier) > OUTPUT_FREQ_MAX) ) {
    	return FALSE;
    }

    return TRUE;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_InitClocks
//* \brief This function performs very low level HW initialization
//*----------------------------------------------------------------------------
unsigned char AT91F_InitClocks(int PLLAR_Register,int PLLBR_Register ,int MCKR_Register)
{
#ifdef  DELAY_ON
    volatile char 	tmp = 0;
#endif
    unsigned int	MainClock;
    unsigned int 	pllDivider,pllMultiplier;

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Optionnal
/////////////////////////////////////////////////////////////////////////////////////////////////////

//* Check if Input & Output Frequencies are in the correct range

    MainClock 		= AT91F_CKGR_GetMainClock(AT91C_BASE_CKGR,SLOWCLOCK);

    pllDivider    	= (PLLAR_Register  & AT91C_CKGR_DIVA);
    pllMultiplier 	= ((PLLAR_Register  & AT91C_CKGR_MULA) >> 16) + 1;
    if(AT91F_CheckPLL_FrequencyRange(MainClock, pllDivider , pllMultiplier) == FALSE) {
	return FALSE;
    }

    pllDivider    	= (PLLBR_Register  & AT91C_CKGR_DIVB);
    pllMultiplier 	= ((PLLBR_Register  & AT91C_CKGR_MULB) >> 16) + 1;
    if(AT91F_CheckPLL_FrequencyRange(MainClock, pllDivider , pllMultiplier) == FALSE) {
	return FALSE;
    }

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Setting PLLA and Divider A
/////////////////////////////////////////////////////////////////////////////////////////////////////

    AT91C_BASE_CKGR->CKGR_PLLAR = PLLAR_Register;
//* Wait for PLLA stabilization LOCKA bit in PMC_SR
#ifdef  DELAY_ON
    tmp = 0;
#endif 								/* DELAY_ON */
    while( !(AT91C_BASE_PMC->PMC_SR & AT91C_PMC_LOCKA) 
#ifdef  DELAY_ON
    	&& (tmp++ < DELAY_PLL) 
#endif 								/* DELAY_ON */
    ) { ; }

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Setting PLLB and Divider B
/////////////////////////////////////////////////////////////////////////////////////////////////////

    AT91C_BASE_CKGR->CKGR_PLLBR = PLLBR_Register;
//* Wait for PLLB stabilization LOCKB bit in PMC_SR
#ifdef  DELAY_ON
    tmp = 0;
#endif
    while( !(AT91C_BASE_PMC->PMC_SR & AT91C_PMC_LOCKB) 
#ifdef  DELAY_ON
    	&& (tmp++ < DELAY_PLL) 
#endif 
    ) { ; }

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Selection of Master Clock MCK (and Processor Clock PCK)
/////////////////////////////////////////////////////////////////////////////////////////////////////

    //* Constraints of the Master Clock selection sequence
    //* Write in the MCKR dirty value concerning the clock selection CSS then overwrite it in a second sequence
    AT91C_BASE_PMC->PMC_MCKR = AT91C_PMC_CSS_SLOW_CLK;
    //* Wait until the master clock is established
#ifdef  DELAY_ON
    tmp = 0;
#endif
    while( !(AT91C_BASE_PMC->PMC_SR & AT91C_PMC_MCKRDY)
#ifdef  DELAY_ON
	 && (tmp++ < DELAY_MAIN_FREQ )
#endif
     );

    //* Second sequence
    AT91C_BASE_PMC->PMC_MCKR = MCKR_Register;
    //* Wait until the master clock is established
#ifdef  DELAY_ON
    tmp = 0;
#endif
    while( !(AT91C_BASE_PMC->PMC_SR & AT91C_PMC_MCKRDY)
#ifdef  DELAY_ON
 	&& (tmp++ < DELAY_MAIN_FREQ)
#endif
    );

    return TRUE;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_InitSDRAM
//* \brief This function performs very low level HW initialisation
//*----------------------------------------------------------------------------
void AT91F_InitSDRAM(void)
{
    volatile int *pSDRAM = (int *) BASE_EBI_CS1_ADDRESS;
	int 	sz;								// 0 - 32Mb, 1 - 16Mb

   	//* Configure PIOC as peripheral (D16/D31)
   	AT91F_SDRC_CfgPIO();

/* uncomment to enable pullup DATA bus */
//#define XDATA_PULLUP

	for(sz = 0; sz < 2; ++sz) { 
		extern unsigned long sdram_size;
		unsigned long pioc_val;

    	//* Setup MEMC to support CS1=SDRAM
    	AT91C_BASE_EBI->EBI_CSA |= AT91C_EBI_CS1A;
    	AT91C_BASE_EBI->EBI_CFGR =

#ifdef XDATA_PULLUP
    	 (AT91C_EBI_DBPUC & 0x00) |		// 0 - pulled up
#else 
    	 (AT91C_EBI_DBPUC) | 			// 1 - not pulled up
#endif 													/* XDATA_PULLUP */
    	 (AT91C_EBI_EBSEN & 0x00);
		
		pioc_val =
	   (((unsigned int)AT91C_PC20_D20) |
		((unsigned int)AT91C_PC21_D21) |
		((unsigned int)AT91C_PC30_D30) |
		((unsigned int)AT91C_PC22_D22) |
		((unsigned int)AT91C_PC31_D31) |
		((unsigned int)AT91C_PC23_D23) |
		((unsigned int)AT91C_PC16_D16) |
		((unsigned int)AT91C_PC24_D24) |
		((unsigned int)AT91C_PC17_D17) |
		((unsigned int)AT91C_PC25_D25) |
		((unsigned int)AT91C_PC18_D18) |
		((unsigned int)AT91C_PC26_D26) |
		((unsigned int)AT91C_PC19_D19) |
		((unsigned int)AT91C_PC27_D27) |
		((unsigned int)AT91C_PC28_D28) |
		((unsigned int)AT91C_PC29_D29));

#ifdef XDATA_PULLUP
	AT91C_BASE_PIOC->PIO_PPUER |= pioc_val; // enable pullup
#else 
	AT91C_BASE_PIOC->PIO_PPUDR |= pioc_val; // disable pullup
#endif 										/* XDATA_PULLUP */

		{
			unsigned long val; 

			sdram_clk = val = 1000000 / ((18432 / PLLA_DIV) * (PLLA_MULA + 1) / (MCKR_MDIV + 1));

#define AT91C_SDRC_TWR_VAL  20 // 1CLK+7ns 10ns (SDRC) Number of Write Recovery Time Cycles
#define AT91C_SDRC_TRC_VAL  70 // 70 ns (SDRC) Number of RAS Cycle Time Cycles
#define AT91C_SDRC_TRP_VAL  20 // 20 ns (SDRC) Number of RAS Precharge Time Cycles
#define AT91C_SDRC_TRCD_VAL 20 // 20 ns (SDRC) Number of RAS to CAS Delay Cycles
#define AT91C_SDRC_TRAS_VAL 50 // 50 ns(SDRC) Number of RAS Active Time Cycles
#define AT91C_SDRC_TXSR_VAL 80 // 80 (SDRC) Number of Command Recovery Time Cycles

			sdram_twr  = (AT91C_SDRC_TWR_VAL  + (val - 1)) / val;
			sdram_trc  = (AT91C_SDRC_TRC_VAL  + (val - 1)) / val;
			sdram_trp  = (AT91C_SDRC_TRP_VAL  + (val - 1)) / val;
			sdram_trcd = (AT91C_SDRC_TRCD_VAL + (val - 1)) / val;
			sdram_tras = (AT91C_SDRC_TRAS_VAL + (val - 1)) / val;
			sdram_txsr = (AT91C_SDRC_TXSR_VAL + (val - 1)) / val;

			val  = 
    			AT91C_SDRC_NR_12 			|
    			AT91C_SDRC_NB_4_BANKS		|
    			AT91C_SDRC_CAS_2			|
				((sdram_twr  & 0xF) <<  7)  |
				((sdram_trc  & 0xF) << 11)	|
				((sdram_trp  & 0xF) << 15)	|
				((sdram_trcd & 0xF) << 19)	|
				((sdram_tras & 0xF) << 23)	|
				((sdram_txsr & 0xF) << 27);

			if(sz == 0) {
		// 32Mb, 9 column 
				sdram_size = (0x100000UL * 32);
    			AT91C_BASE_SDRC->SDRC_CR = (AT91C_SDRC_NC_9 | val);
			} else {
		// 16Mb, 8 column
				sdram_size = (0x100000UL * 16);
    			AT91C_BASE_SDRC->SDRC_CR = (AT91C_SDRC_NC_8 | val);
			}
		}

    //* A Precharge All command is issued to the SDRAM
    	AT91C_BASE_SDRC->SDRC_MR = AT91C_SDRC_MODE_PRCGALL_CMD;
    	*pSDRAM = 0;

    //* Eight Auto-refresh are provided
    	AT91C_BASE_SDRC->SDRC_MR = AT91C_SDRC_MODE_RFSH_CMD;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;
		*pSDRAM = 0;

    //* A mode register cycle is issued to program the SDRAM parameters
    	AT91C_BASE_SDRC->SDRC_MR = AT91C_SDRC_MODE_LMR_CMD;
    	*(pSDRAM + 0x0020) = 0;

    //* Write refresh rate into SDRAMC refresh timer COUNT register
    	AT91C_BASE_SDRC->SDRC_TR = SDRC_TR_VAL; /* (AT91C_SDRC_COUNT & 0x2E0 ) */ /* 936 */;
    	*pSDRAM = 0;

    //* A Normal Mode Command is provided, 3 clocks after tMRD is set
    	AT91C_BASE_SDRC->SDRC_MR = (AT91C_SDRC_MODE_NORMAL_CMD | AT91C_SDRC_DBW_32_BITS);
    	*pSDRAM = 0;


		{
			int pos;

			for(pos = 0; pos < sdram_size / sizeof(int) ; ) {
    			*(pSDRAM + pos) = pos;
				if(pos == 0) {
					pos = 1;
				} else {
					pos <<= 1;
				}
			}

			for(pos = 0; pos < sdram_size / sizeof(int); ) {
    			if( *(pSDRAM + pos) != pos ) {
					break;
    			}
				if(pos == 0) {
					pos = 1;
				} else {
					pos <<= 1;
				}
			}

			if( pos >= ( sdram_size / sizeof(int) ) ) {
				break;
			}
		}
	}
}

void AT91F_InitNand(void)
{
    /* Setup Smart Media, first enable the address range of CS3 */
    AT91C_BASE_EBI->EBI_CSA |= AT91C_EBI_CS3A_SMC_SmartMedia;

    /* set the bus interface characteristics based on
       tDS Data Set up Time 30 - ns
       tDH Data Hold Time 20 - ns
       tALS ALE Set up Time 20 - ns
       16ns at 60 MHz ~= 3  */
#define AT91C_SM_ID_RWH	(5 << 28)		/* debug only orig = 5 */
#define AT91C_SM_RWH	(1 << 28)		/* orig = 1 */
#define AT91C_SM_RWS	(0 << 24)		/* orig = 0 */
#define AT91C_SM_TDF	(1 << 8)		/* orig = 1 */
#define AT91C_SM_NWS	(3)			/* orig = 3 */
    AT91C_BASE_SMC2->SMC2_CSR[3] =
      ( AT91C_SM_RWH | AT91C_SM_RWS |
	AT91C_SMC2_ACSS_STANDARD |
	AT91C_SMC2_DBW_8 | AT91C_SM_TDF |
	AT91C_SMC2_WSEN | AT91C_SM_NWS);

    /* Enable NAND Flash (SmartMedia) interface. */
    /* enable PC0=SMCE, PC1=SMOE, PC3=SMWE, A21=CLE, A22=ALE */
    /* PC1 - SMOE peripherial A, PC3 - SMWE peripherial A */ 
    AT91C_BASE_PIOC->PIO_ASR = (AT91C_PC1_BFRDY_SMOE | AT91C_PC3_BFBAA_SMWE);
    AT91C_BASE_PIOC->PIO_PDR = (AT91C_PC1_BFRDY_SMOE | AT91C_PC3_BFBAA_SMWE);

    /* PC5 - output, 0 */
    AT91C_BASE_PIOC->PIO_PER = AT91C_PC5_BFWE;	/* PIO enable */
    AT91C_BASE_PIOC->PIO_OER = AT91C_PC5_BFWE;	/* output enable */
    AT91C_BASE_PIOC->PIO_SODR= AT91C_PC5_BFWE;	/* set output data register */
    AT91C_BASE_PIOC->PIO_OWER= AT91C_PC5_BFWE;	/* output write enable register */

    /* Configure PC2 as input (signal READY of the SmartMedia) */
    AT91C_BASE_PIOC->PIO_PER = AT91C_PC2_BFAVD;	/* enable direct output enable */
    AT91C_BASE_PIOC->PIO_ODR = AT91C_PC2_BFAVD;	/* disable output */

    /* Configure PB1 as input (signal Card Detect of the SmartMedia) */
    AT91C_BASE_PIOB->PIO_PER = AT91C_PIO_PB1;	/* enable direct output enable */
    AT91C_BASE_PIOB->PIO_ODR = AT91C_PIO_PB1;	/* disable output */
}

// static inline int AT91_PIO_SmartMedia_CardDetect(void) {
// 	return (AT91_SYS->PIOB_PDSR & AT91C_PIO_PB1) ? 1 : 0;
// }



//*----------------------------------------------------------------------------
//* \fn    AT91F_InitFlash
//* \brief This function performs very low level HW initialization
//*----------------------------------------------------------------------------
void AT91F_InitFlash(void)
{
    AT91C_BASE_MC->MC_PUIA[0] = AT91C_MC_PROT_PNAUNA;
    AT91C_BASE_MC->MC_PUP = 0;
    AT91C_BASE_MC->MC_PUER = 0;	//* Memory controller protection unit disable
    AT91C_BASE_MC->MC_ASR = 0;  //* read only!
    AT91C_BASE_MC->MC_AASR = 0; //* read only!

    //* Setup MEMC to support CS0=Flash
    AT91C_BASE_EBI->EBI_CSA |= AT91C_EBI_CS0A_SMC;
    AT91C_BASE_EBI->EBI_CFGR = 
#ifdef XDATA_PULLUP
    (AT91C_EBI_DBPUC & 0x00)|	// 0 - pulled up
#else 
    (AT91C_EBI_DBPUC ) 		| 	// 1 - not pulled up
#endif 													/* XDATA_PULLUP */
    (AT91C_EBI_EBSEN & 0x00);

    //* Setup Flash
#define SMC_CSR_VAL 				\
(						\
    (AT91C_SMC2_NWS & 0x07 ) 		|	\
     AT91C_SMC2_WSEN     		|	\
    (AT91C_SMC2_TDF & 0x200) 		|	\
     AT91C_SMC2_DBW_16				\
)

    AT91C_BASE_SMC2->SMC2_CSR[0] = SMC_CSR_VAL;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_LowLevelInit
//* \brief This function performs very low level HW initialization
//*----------------------------------------------------------------------------
void AT91F_LowLevelInit(void)
{
    unsigned char status;

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Flash Initialisation : must be done before change the clocks -> Set correct Wait States
/////////////////////////////////////////////////////////////////////////////////////////////////////
    AT91F_InitFlash();
    AT91F_InitNand();
 
/////////////////////////////////////////////////////////////////////////////////////////////////////
    status = AT91F_InitClocks(PLLAR,PLLBR,MCKR);

//  AIC init code moved from "C" code to startup code  
	
// Open PIO for DBGU
    AT91F_DBGU_CfgPIO();

	{
		/* */
		master_clock = ((18432000 / PLLA_DIV) * (PLLA_MULA + 1) / (MCKR_MDIV + 1));
	}

// Configure DBGU
    AT91F_US_Configure (
	(AT91PS_USART) AT91C_BASE_DBGU,          	// DBGU base address
#ifdef OLD_CLK
	(179712000 / 3),
#else 
	 master_clock,
#endif 														/* OLD_CLK */
	(AT91C_US_CHMODE_NORMAL | AT91C_US_PAR_NONE) ,  // mode Register to be programmed
	 115200 ,        								// baudrate to be programmed
	0);                        						// timeguard to be programmed


// Enable Transmitter
    AT91F_US_EnableTx((AT91PS_USART) AT91C_BASE_DBGU);

// Enable Receiver
    AT91F_US_EnableRx((AT91PS_USART) AT91C_BASE_DBGU);

/////////////////////////////////////////////////////////////////////////////////////////////////////
// SDRAM Initialisation
/////////////////////////////////////////////////////////////////////////////////////////////////////
    AT91F_InitSDRAM();


#ifdef OLD_CLK
    AT91F_DBGU_Printk("\n59.904MHz\n");
#else 
#if !defined SPIBOOT && !defined NANDWRITE
	{
		AT91F_DBGU_Printv("\nMaster clock", master_clock, "Hz\n");
		AT91F_DBGU_Printv("twr" , sdram_twr , 0);
		AT91F_DBGU_Printv("trc" , sdram_trc , 0);
		AT91F_DBGU_Printv("trp" , sdram_trp , 0);
		AT91F_DBGU_Printv("trcd", sdram_trcd, 0);
		AT91F_DBGU_Printv("tras", sdram_tras, 0);
		AT91F_DBGU_Printv("txsr", sdram_txsr, 0);
		AT91F_DBGU_Printv("tr " , SDRC_TR_VAL, 0);
		AT91F_DBGU_Printv("clk ", sdram_clk , 0);
	}
#endif 							/* SPIBOOT */
#endif 
}
