//*----------------------------------------------------------------------------
//* ATMEL Microcontroller Software Support - ROUSSET -
//*----------------------------------------------------------------------------
//* The software is delivered "AS IS" without warranty or condition of any
//* kind, either express, implied or statutory. This includes without
//* limitation any warranty or condition with respect to merchantability or
//* fitness for any particular purpose, or against the infringements of
//* intellectual property rights of others.
//*----------------------------------------------------------------------------
//* File Name           : main.c
//* Object              : main application written in C
//* Creation            : FB   21st March 2003
//*
//*----------------------------------------------------------------------------

/* modified 2004-2007 by Dmitriy Cherkashin */
/* email: divch@users.sourceforge.net, dch@ucrouter.ru */
/* site: www.ucrouter.ru phone: +7-910-422-2917 */
/* based on ATMEL AT91RM9200-BasicBoot */
/* and AT91RM9200-BasicROM_Services code */
/* MIT licence */

#include "AT91RM9200.h"
#include "embedded_services.h"

#ifdef __GNUC__
#include "proto.h"			
#else 
#include "lib_AT91RM9200.h"
#endif							/* __GNUC__ */

#define AT91RM9200_29LV160D_VERSION	"1.1"

#define FILE_MAX_SIZE (15 * 1024 * 1024) /* file max size */

/* */
#define AT91_SMARTMEDIA_BASE 0x40000000	/* */
#define AT91_SMART_MEDIA_ALE    (1<<22)	/* ALE */
#define AT91_SMART_MEDIA_CLE    (1<<21)	/* CLE */
#define NAND_K9F1G08U0A	         0xECF1	/* nand flash id */
static unsigned short nand_id; 		/* nand flash id */
static unsigned char nand_size;		/* nand flash size (nand_size**2) */
static unsigned char nand_page;		/* nand flash page size (nand_page**2) */
static unsigned char nand_oob;		/* nand oob size (nand_oob**2) */
static unsigned char nand_block;	/* erase block size (nand_block**2) */
// TODO static volatile unsigned char * nand_cle = (volatile unsigned char *) (AT91_SMARTMEDIA_BASE | AT91_SMART_MEDIA_CLE);
// TODO static volatile unsigned char * nand_ale = (volatile unsigned char *) (AT91_SMARTMEDIA_BASE | AT91_SMART_MEDIA_ALE);

#define SPI_BOOT_SIZE			    (16*1056)
#define AT91C_SPI_PCS0_SERIAL_DATAFLASH		0xE   // Chip Select 0 : NPCS0 %1110
#define AT91C_DATAFLASH_TIMEOUT		      10000   // Timeout for AT91F_DataFlashWaitReady

#define SDRAM_ADDR   		        (0x20000000)  // SDRAM bank address
#define FLASH_ADDR   		        (0x10000000)  // CS0 address
unsigned long sdram_size = 0; 		  			  // SDRAM SIZE
unsigned long master_clock = 0;
/* comment next definition (and modify CS0 bus width settings in init.c)
 to program flash in 8 bit mode 
*/
#define FLASH_16							


////////////////////////////////////////////////////////////////////////////////
// External Functions
////////////////////////////////////////////////////////////////////////////////
extern void AT91F_DBGU_Printk(char *);			  // print string value 
extern void AT91F_DBGU_Printl(unsigned long val); // print long value  
extern void AT91F_DBGU_Printx(unsigned long val); // print hexadecimal value
extern void AT91F_DBGU_Printv(char * pref, unsigned long val, char *suf);
extern void AT91F_DBGU_Printw(char * pref, unsigned long val, char *suf);
extern int AT91F_DBGU_getch(char * ch);			  // read character from console
extern void AT91F_DBGU_getx(unsigned long * val); // read hexadecimal value from console
extern void AT91F_ASM_SPI_Handler(void);		  // SPI handler 
extern void AT91F_ASM_ST_DBGU_Handler(void);	  // ST/DBGU handler 

////////////////////////////////////////////////////////////////////////////////
// Allocation of the service structures
////////////////////////////////////////////////////////////////////////////////

// "ROM Entry" Service
AT91S_RomBoot const *pAT91;

// "SBuffer and Pipe" structures for "Xmodem" Service
AT91S_SBuffer	sXmBuffer;
AT91S_Pipe	xmodemPipe;
AT91S_SvcXmodem svcXmodem;

// "Tempo" Service
AT91S_CtlTempo  ctlTempo;

// "DataFlash" Service
AT91S_SvcDataFlash 	svcDataFlash;
AT91S_Dataflash		DeviceAT45DB;

////////////////////////////////////////////////////////////////////////////////
// Global Variables
////////////////////////////////////////////////////////////////////////////////
unsigned char		  * Buffer_Dataflash = 0;
volatile unsigned int 	StTick;					// system timer tick count for timeout 
volatile char			EndOfTransmission = 0; 	// flag for end of xmodem

#ifndef NANDWRITE
//*----------------------------------------------------------------------------
//* \fn    AT91F_DataFlashWaitReady
//* \brief wait for dataflash ready (bit7 of the status register == 1)
//*----------------------------------------------------------------------------


AT91S_SvcDataFlashStatus AT91F_DataFlashWaitReady(unsigned int timeout)
{
	unsigned int 	i;

	svcDataFlash.DataFlashDesc.DataFlash_state = IDLE;	

	do
	{
		svcDataFlash.Status(&(svcDataFlash.DataFlashDesc));
		timeout--;
		// dummy waiting time
		for(i=0;i<10;i++);
	}
	while( ((svcDataFlash.DataFlashDesc.DataFlash_state & 0x80) != 0x80) && (timeout>0) );

	if((svcDataFlash.DataFlashDesc.DataFlash_state & 0x80) != 0x80)
		return DATAFLASH_ERROR;
		
	return DATAFLASH_OK;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_SPI_CfgSPI
//* \brief Config SPI IP
//*----------------------------------------------------------------------------
void  AT91F_SPI_CfgSPI(void)
{
	// Reset the SPI
    AT91F_SPI_Reset(AT91C_BASE_SPI);

    // Configure SPI in Master Mode with No CS selected !!!  
    AT91F_SPI_CfgMode(AT91C_BASE_SPI, AT91C_SPI_MSTR | AT91C_SPI_MODFDIS | AT91C_SPI_PCS);

	// Configure SPI CS0 for Serial svcDataFlash AT45DBxx
    AT91F_SPI_CfgCs(0, AT91C_SPI_CPOL | (AT91C_SPI_DLYBS & 0x100000) | ((60*1000*1000 / (2*DATAFLASH_CLK)) << 8));
//	AT91F_SPI_CfgCs(0, AT91C_SPI_CPOL | (AT91C_SPI_DLYBS & 0x100000) | ((master_clock / (2*DATAFLASH_CLK)) << 8));

    // Enable the SPI
    AT91F_SPI_Enable(AT91C_BASE_SPI);
}
#endif 							/* NANDWRITE */


//*----------------------------------------------------------------------------
//* \fn    AT91_XmodemComplete
//* \brief Allows to correctly complete xmodem end of transmission
//*----------------------------------------------------------------------------
void AT91_XmodemComplete(AT91S_PipeStatus status, void *pVoid)
{
// stop the Xmodem tempo 
    svcXmodem.tempo.Stop(&(svcXmodem.tempo));
	
// upload complete
    EndOfTransmission = 1;
}

//*----------------------------------------------------------------------------
//* \fn    AT91_XmodemProtocol
//* \brief Allows to correctly complete xmodem end of transmission
//*----------------------------------------------------------------------------
void AT91_XmodemProtocol(AT91S_PipeStatus status, void *pVoid)
{
    AT91PS_USART  pUsart     = svcXmodem.pUsart;
			
    svcXmodem.tempo.Start(&(svcXmodem.tempo), 10, 0, AT91_XmodemComplete, pUsart);								
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_SPI_Handler
//* \brief SPI Handler
//*----------------------------------------------------------------------------
void AT91F_SPI_Handler(void)
{
    int status;

    status = ( AT91C_BASE_SPI->SPI_SR & AT91C_BASE_SPI->SPI_IMR );

    svcDataFlash.Handler(&svcDataFlash, status);
}


///////////////////////////////////////////////////////////////////////////////////////////
// Interrupt handlers
///////////////////////////////////////////////////////////////////////////////////////////

//*----------------------------------------------------------------------------
//* \fn    AT91F_ST_DBGU_Handler
//* \brief C Interrupt handler for Interrupt source 1
//*----------------------------------------------------------------------------
void AT91F_ST_DBGU_Handler(void)
{
    volatile unsigned int csr = AT91C_BASE_DBGU->DBGU_CSR;

    // ========== Systimer interrupt ============== //
    if (AT91C_BASE_ST->ST_SR & AT91C_ST_PITS) {
    	StTick++;
	ctlTempo.CtlTempoTick(&ctlTempo);
	return;
    }

    // ========== DBGU interrupt ============== //
    if (csr & (AT91C_US_PARE  | AT91C_US_FRAME | AT91C_US_OVRE | AT91C_US_RXBRK)) 
    {
// Stop previous Xmodem transmission
	AT91C_BASE_DBGU->DBGU_CR = AT91C_US_RSTSTA;
	AT91C_BASE_DBGU->DBGU_IDR = AT91C_US_ENDRX;
	AT91C_BASE_DBGU->DBGU_IER = AT91C_US_RXRDY | AT91C_US_COMM_RX;	
    } else if (csr & ((AT91C_US_TXRDY | AT91C_US_ENDTX | AT91C_US_TXEMPTY) | 
	   (AT91C_US_RXRDY | AT91C_US_ENDRX | AT91C_US_TIMEOUT) | 
	   AT91C_US_RXBUFF))
    {
        svcXmodem.Handler(&svcXmodem, csr);
    }
}

/* SDRAM test */

int sdram_test(int mode) 
{
 int passnum, errpass;   
 int rvl = 1;
 passnum = 0;
 errpass = 0;
 while(1) {
    unsigned long * start = (unsigned long *) ( SDRAM_ADDR );
    unsigned long * end  =  (unsigned long *) ( SDRAM_ADDR + sdram_size);
    unsigned long * addr;
    unsigned long rval;
    unsigned long val;
	
    AT91F_DBGU_Printv("SDRAM test (", sdram_size, " bytes)");
    if(mode == 1) {
	AT91F_DBGU_Printv(" pass "    , passnum, "");
	AT91F_DBGU_Printv(", errorrs ", errpass, "");
    }
    AT91F_DBGU_Printk("\n");

    addr = start;
    val  = 1;
    while(addr < end) {
	*addr++ = val++;
    }

    addr = start;
    val  = 1;
    while(addr < end) {
        rval = *addr;
        if(rval != val) {
    	    AT91F_DBGU_Printw("Error, offset ", (unsigned long) addr, "" );
	    AT91F_DBGU_Printw("writen ", val,  "" );
	    AT91F_DBGU_Printw("readed ",rval,"\n" );
	    break;
	}
	addr++;
	val++;
    }

    if(addr < end) {
        errpass++;
        rvl = 0;
    } else {
        AT91F_DBGU_Printk("Ok\n");
    }
    passnum++;
			
    if(mode == 0) {
        break;
    } else {
	char it;
	it = 0;
	if(AT91F_DBGU_getch( & it )) {
	    if(it == 'q' || it == 'Q') {
		break;
	    }
	}
    }
 }
 return(rvl);
}

#ifdef NANDWRITE

/* nand flash function */
#define NAND_CMD_READ		0	/* read start */
#define NAND_CMD_READ_END	0x30	/* read end */ 
#define NAND_CMD_WRITE		0x80	/* write start */
#define NAND_CMD_WRITE_END	0x10	/* write end */
#define NAND_CMD_ERASE		0x60	/* erase start */
#define NAND_CMD_ERASE_END	0xd0	/* erase end */
#define NAND_CMD_STATUS		0x70	/* status */
#define NAND_CMD_READID		0x90	/* read ID */ 
#define NAND_CMD_RESET		0xff	/* reset */

/* return ready flag */
static int nand_ready(void)
{
    return((AT91C_BASE_PIOC->PIO_PDSR & AT91C_PIO_PC2));
}

/* read byte */
static unsigned char nand_read_byte(void)
{
    return(*((volatile unsigned char *) AT91_SMARTMEDIA_BASE));
}

/* write byte */
static void nand_write_byte(unsigned char val)
{
    *((volatile unsigned char *)AT91_SMARTMEDIA_BASE) = val; 
}

static void nand_write_cmd(unsigned char val)
{
     *((volatile unsigned char *)(AT91_SMARTMEDIA_BASE | AT91_SMART_MEDIA_CLE)) = val; 
}

static void nand_write_addr(unsigned char val)
{
    *((volatile unsigned char *)(AT91_SMARTMEDIA_BASE | AT91_SMART_MEDIA_ALE)) = val; 
}

/* select nand flash */
static int nand_select(void)
{
    AT91C_BASE_PIOC->PIO_CODR = AT91C_PC5_BFWE;	/* clear PC5 - NCE */
}

/* deselect nand flash */
static int nand_deselect(void)
{
    AT91C_BASE_PIOC->PIO_SODR = AT91C_PC5_BFWE;	/* set PC5 */
}

/* wait ready & read status */
static int nand_wait(void)
{
 unsigned char status;
 while(1) {
//    if(nand_ready()) {
//	break;
//    }
//
    nand_select();
    nand_write_cmd(NAND_CMD_STATUS);	/* read status */
    status = nand_read_byte();
    if((status & 0x40) != 0) {
	break;
    }
    nand_deselect();
 }
 nand_select();
 nand_write_cmd(NAND_CMD_STATUS);
 status = nand_read_byte();		/* read status */
 nand_deselect();			/* deselect nand flash */
 return(status);			/* return status */
}

/* nand flash read, col = (-1) read obb  */
int nand_read(int row, int col, unsigned char * buf, int len)
{
 int psize;
 int pos;

 if(buf == 0) {
    return(0);
 }

 psize = (0x1 << nand_page);		/* page size */	
 if(col == -1) {
 /* read oob data */    
    int osize;
    osize = (0x1 << nand_oob);		/* oob size */
    if(len < 0) {			/* default oob buffer length */ 
	len = osize;
    }
    col = psize;
    psize += osize;
 } else {
    if(col >= psize) {
	return(0);
    }
 }

 nand_select();				/* select nand flash */
 nand_write_cmd(NAND_CMD_READ);		/* read command */
 nand_write_addr((unsigned char) (col & 0xFF));		/* col1 */
 nand_write_addr((unsigned char) ((col >> 8) & 0xFF));	/* col2 */
 nand_write_addr((unsigned char) (row & 0xff));		/* row1 */
 nand_write_addr((unsigned char) ((row >> 8) & 0xff));	/* row2 */
 nand_write_cmd(NAND_CMD_READ_END);	/* read end */
 while(nand_ready() == 0) {;} 

 pos = 0;
 while(pos < len && col < psize) {
    *(buf + pos) = nand_read_byte();
    pos++;
    col++;
 }
 nand_deselect();			/* deselect nand flash */
 return(pos);				/* readed bytes */
}


/* nand flash write, col = (-1), write oob data */
int nand_write(int row, int col, unsigned char * buf, int len)
{
 int status;				/* status */
 int psize;				/* page size */
 int pos;

 if(buf == 0) {
    return(0);
 }

 psize = (0x1 << nand_page);
 if(col == -1) {
    /* write oob data */
    int osize;
    osize = (0x1 << nand_oob);
    if(len < 0) {
	len = osize;
    }
    col = psize;
    psize += osize;
 } else {
    if(col >= psize) {
	return(0);
    }
 }

 nand_select();				/* select nand flash */
 nand_write_cmd(NAND_CMD_WRITE);	/* start program */
 nand_write_addr((unsigned char)  (col & 0xFF));	/* col1 */
 nand_write_addr((unsigned char) ((col >> 8) & 0xFF));	/* col2 */
 nand_write_addr((unsigned char)  (row & 0xFF));	/* row1 */
 nand_write_addr((unsigned char) ((row >> 8) & 0xFF));	/* row2 */

 pos = 0;
 while(pos < len && col < psize) {
    nand_write_byte(*(buf + pos));
    pos++;
    col++;
 }
 nand_write_cmd(NAND_CMD_WRITE_END);	/* program end command*/
 nand_deselect();

 status = nand_wait();
 if(status & 0x01) {
    return(-1);
 } else {
    return(pos);			/* bytes programmed */
 }
}

static int nand_erase(int row)
{
 int status;
 row = row & (((unsigned int) -1) << (nand_block - nand_page));
 nand_select();					/* select nand flash */
 nand_write_cmd(NAND_CMD_ERASE);		/* erase command */
 nand_write_addr((unsigned char) (row & 0xff));		/* row1 */
 nand_write_addr((unsigned char) ((row >> 8) & 0xff));	/* row2 */
 nand_write_cmd(NAND_CMD_ERASE_END);		/* erase end command */
 nand_deselect();

 status = nand_wait();
 if(status & 0x01) {
    return(-1);
 } else {
    return(1);
 }
}

/* detect Nand flash */
int nand_detect(void)
{
    unsigned char man;
    unsigned char dev;

    AT91F_DBGU_Printk("Init NAND flash ");

    nand_select();
    nand_write_cmd(NAND_CMD_READID);		/* READID */
    nand_write_addr(0x00);
    nand_write_addr(0x00);

    man = nand_read_byte();
    dev = nand_read_byte();
    nand_deselect();

    if(man == 0xEC && dev == 0xF1) {
	AT91F_DBGU_Printk("K9F1G08U0A detected\n");
	nand_id = NAND_K9F1G08U0A; 
	nand_size = 27;			/* (1 << 27) = 0x08000000 128M */
	nand_block = 17;		/* (1 << 17) = 0x20000 128K */
	nand_page = 11;			/* (1 << 11) = 0x0800 2K */
	nand_oob = 6;			/* (1 <<  6) = 40 64 */
    } else {
	AT91F_DBGU_Printw("Unknown NAND flash\nmanufacture ",man,"");
	AT91F_DBGU_Printw("device ",dev,"");
	nand_id = 0; 
	nand_size = 0;
	nand_block = 0;
	nand_page = 0;
	nand_oob = 0;
    }

    nand_select();
    nand_write_cmd(NAND_CMD_READID);		/* READID */
    nand_write_addr(0x00);
    nand_write_addr(0x00);

    man = nand_read_byte();
    dev = nand_read_byte();
    nand_deselect();

#if 0
    AT91F_DBGU_Printw("size" ,(1 << nand_size),"");
    AT91F_DBGU_Printw("page" ,(1 << nand_page),"");
    AT91F_DBGU_Printw("oob"  ,(1 << nand_oob),"");
    AT91F_DBGU_Printw("block",(1 << nand_block),"");
#endif /* 0 */

    return(1);
}
#endif							/* NANDWRITE */


int AT91_dump(void * addr, int cnt)
{
 int off;
 int row;
 if(addr < 0) {
    return(0);
 } 

 /* dump cnt bytes */
 off = 0;
 for(row = 0; row < 16 && cnt > 0; ++row) {
    int col;
    for(col = 0; col < 16 && cnt > 0; ++col, ++off) {
	unsigned char ch;
	char buf[3];
	ch = *((unsigned char *)addr + off);
	ch = (ch >> 4) & 0xF;
	if(/*ch >= 0 && */ ch <= 9) {
	    ch = '0' + ch;
	} else {
	    ch = 'A' + (ch - 10);
	}
	buf[0] = ch;
	
	ch = *((unsigned char *)addr + off);
	ch = (ch & 0xF);
	if(/* ch >= 0 && */ ch <= 9) {
	    ch = '0' + ch;
	} else {
	    ch = 'A' + (ch - 10);
	}
	
	buf[1] = ch;
	buf[2] = '\0';
	cnt--;
				
	AT91F_DBGU_Printk( buf );
	if( ( (col + 1) % 4) == 0) {
	    AT91F_DBGU_Printk(" ");
	}
    }	
		    
    AT91F_DBGU_Printk("\n");
 }

 AT91F_DBGU_Printk("[Q]uit\n");

 while(1) {
    unsigned char it;
    it = 0;
    if(AT91F_DBGU_getch( & it )) {
	if(it == 'q' || it == 'Q') {
	    return(0);
	} else {
	    return(1);
	}
    }
 }
}  


//*----------------------------------------------------------------------------
//* \fn    main
//* \brief main function
//*----------------------------------------------------------------------------
int main(void)
{
 AT91PS_Buffer  pXmBuffer;
 AT91PS_SvcComm pSvcXmodem;
 unsigned char man_Id    = 0;	// man ID 
 unsigned char device_Id = 0;	// device ID
 int stp;

#ifndef SPIBOOT
    AT91F_DBGU_Printk("\nAT91RM9200 AM29LV160D Flash Write Programm "
	AT91RM9200_29LV160D_VERSION
	"\n"
    );
#endif 							/* SPIBOOT */
 {								
    sdram_test(0); 
 }

 // Call Open methods:
 pAT91 = AT91C_ROM_BOOT_ADDRESS;

///////////////////////////////////////////////////////////////////////////////////////////
//  Init of the service Tempo
///////////////////////////////////////////////////////////////////////////////////////////

 // Open Tempo service
 pAT91->OpenCtlTempo(&ctlTempo, (void *) &(pAT91->SYSTIMER_DESC) );
 ctlTempo.CtlTempoStart((void *) &(pAT91->SYSTIMER_DESC) );

///////////////////////////////////////////////////////////////////////////////////////////
//  Init of the service Xmodem
///////////////////////////////////////////////////////////////////////////////////////////

 // Open SBuffer structure 
 pXmBuffer = pAT91->OpenSBuffer(&sXmBuffer);

 // Open Xmodem service
 pSvcXmodem = pAT91->OpenSvcXmodem(&svcXmodem,(AT91PS_USART) AT91C_BASE_DBGU, &ctlTempo);

 // Open Pipe structure
 pAT91->OpenPipe(&xmodemPipe, pSvcXmodem, pXmBuffer);

///////////////////////////////////////////////////////////////////////////////////////////
//  Init of the service svcDataFlash
///////////////////////////////////////////////////////////////////////////////////////////

#ifndef NANDWRITE
 // Init of the SPI
 AT91F_DBGU_Printk("\nInit SPI flash ");

 for(stp = 0; stp < 2; ++stp) {
    unsigned char type;			// SPI flash status register

    Buffer_Dataflash = (unsigned char *) (SDRAM_ADDR + (sdram_size - 0x10000));

    /* Init SPI for svcDataFlash interface */
    AT91F_SPI_CfgPIO();			// configure SPI Pins
    AT91F_SPI_CfgPMC();			// configure PMC
    AT91F_SPI_CfgSPI();					
    AT91F_PDC_Open(AT91C_BASE_PDC_SPI);	// open PDC

    if(stp == 0) {
    /* probe AT45DB642 */
	DeviceAT45DB.pages_number = 8192;	// number of pages 
	DeviceAT45DB.pages_size   = 1056;	// page size (bytes)
	DeviceAT45DB.page_offset  = 11;		// page low bit offset 
	DeviceAT45DB.byte_mask    = 0x700;	
    } else {
    /* probe AT45DB321 */
	DeviceAT45DB.pages_number = 8192;
	DeviceAT45DB.pages_size   = 528;
	DeviceAT45DB.page_offset  = 10;
	DeviceAT45DB.byte_mask    = 0x300;	
    }

    /* Init AT91S_DataflashDesc Structure */
    svcDataFlash.DataFlashDesc.state 	  	= IDLE;
    svcDataFlash.DataFlashDesc.DataFlash_state	= IDLE;

    /* Init AT91S_SvcDataFlash Global Structure */
    svcDataFlash.pDevice = &DeviceAT45DB;

    /* Configure SPI interrupt */
    AT91F_AIC_ConfigureIt(AT91C_BASE_AIC,
	AT91C_ID_SPI, 3 /* AT91C_AIC_PRIOR_HIGHEST */,
	AT91C_AIC_SRCTYPE_INT_LEVEL_SENSITIVE,
	AT91F_ASM_SPI_Handler
    );

    /* Enable SPI interrupt */
    AT91F_AIC_EnableIt(AT91C_BASE_AIC,AT91C_ID_SPI);

    AT91F_SPI_CfgPCS(AT91C_BASE_SPI,AT91C_SPI_PCS0_SERIAL_DATAFLASH);
    /* Open the service */
    pAT91->OpenSvcDataFlash (AT91C_BASE_PMC, &svcDataFlash);

    /* detect flash  */
    svcDataFlash.DataFlashDesc.DataFlash_state = IDLE;	
    svcDataFlash.Status(&(svcDataFlash.DataFlashDesc));
    type = ((svcDataFlash.DataFlashDesc.DataFlash_state >> 2) & 0x0F);
    if(stp == 0) {
	if( (type & 0x0E) == 0x0E ) {
    	    AT91F_DBGU_Printk("AT45DB642B detected\n");
	    break;
	} 
    } else
    if(stp == 1) {
	if(type == 0x0D) {	// 1101
	    AT91F_DBGU_Printk("AT45DB321B detected\n");
	    break;
	}
    }
 }

 if(stp >= 2) {
    AT91F_DBGU_Printk("Unknown SPI flash\n");
 }	
#endif 							/* NANDWRITE */

///////////////////////////////////////////////////////////////////////////////////////////
//  Init the Uploader
///////////////////////////////////////////////////////////////////////////////////////////

 /* Initialize the Interrupt Source 1 for SysTimer and DBGU */
 AT91F_AIC_ConfigureIt(AT91C_BASE_AIC,
    AT91C_ID_SYS, AT91C_AIC_PRIOR_HIGHEST,
    AT91C_AIC_SRCTYPE_INT_LEVEL_SENSITIVE,
    AT91F_ASM_ST_DBGU_Handler
 );
						 
 /* Enable SysTimer and DBGU interrupt */
 AT91F_AIC_EnableIt(AT91C_BASE_AIC, AT91C_ID_SYS);

 {
#ifdef FLASH_16
    volatile unsigned short * flash_addr = (volatile unsigned short *) FLASH_ADDR;
    /* read man Id */
    *(flash_addr + 0x555) = 0xAA;		/* 1 write command */
    *(flash_addr + 0x2AA) = 0x55;		/* 2 write command */
    *(flash_addr + 0x555) = 0x90;		/* 3 write comman */
      man_Id = *(flash_addr + 0);		/* read manufactorel Id */


    /* read device Id */
    *(flash_addr + 0x555) = 0xAA;		/* 1 */
    *(flash_addr + 0x2AA) = 0x55;		/* 2 */
    *(flash_addr + 0x555) = 0x90;		/* 3 */
      device_Id = *(flash_addr + 1);

    *(flash_addr + 0x0)   = 0xF0;		/* write reset command */

    if( (man_Id  & 0xFF) == 1 && ((device_Id & 0xFF) == 0xC4 || (device_Id & 0xFF) == 0x49)) {
	AT91F_DBGU_Printk("AM29LV160D detected\n");
    }

#else
    volatile unsigned char * flash_addr = (volatile unsigned char *) FLASH_ADDR;
    /* read man Id */
    *(flash_addr + 0xAAA) = 0xAA;		/* 1 write command */
    *(flash_addr + 0x555) = 0x55;		/* 2 write command */
    *(flash_addr + 0xAAA) = 0x90;		/* 3 write comman */
      man_Id = *(flash_addr + 0);		/* read manufactorel Id */

    /* read device Id */
    *(flash_addr + 0xAAA) = 0xAA;		/* 1 */
    *(flash_addr + 0x555) = 0x55;		/* 2 */
    *(flash_addr + 0xAAA) = 0x90;		/* 3 */
      device_Id = *(flash_addr + 2);

    *(flash_addr + 0x0)   = 0xF0;		/* write reset command */

    if(man_Id == 1 && device_Id == 0xC4) {
        AT91F_DBGU_Printk("AM29LV160D detected\n");
    }
#endif 							/* FLASH_16 */
 }

#ifdef NANDWRITE
 nand_detect();
#endif							/* NANDWRITE */

//////////////////////////////////////////////////////////////////////////////////////////
//  Enter the uploading by Xmodem -> Service Xmodem + Service Tempo used
///////////////////////////////////////////////////////////////////////////////////////////

 while(1) {
    AT91S_PipeStatus st; 
    char it;

    char * sdram_addr = (char *) SDRAM_ADDR;
    int sz = FILE_MAX_SIZE;		/* file max size */
    int pos;
    for(pos = 0; pos < sz; ++pos) {
        *(sdram_addr + pos) = 0xFF;
    }


#define	CMD_ERASE   	     ((char)'X') /* erase flash */
#define	CMD_PROGRAM_0 	     ((char)'0') /* program flash with offset 0x00000 */
#define	CMD_PROGRAM_1	     ((char)'1') /* program flash with offset 0x10000 */
#define	CMD_PROGRAM_2	     ((char)'2') /* program flash with offset 0x20000 */
#define	CMD_PROGRAM	     ((char)'G') /* program flash with arbitrary offset */
#define	CMD_READ 	     ((char)'R') /* read flash */
#define	CMD_VERIFY_0         ((char)'V') /* verify flash with offset 0x00000 */
#define	CMD_VERIFY_1         ((char)'I') /* verify flash with offset 0x10000 */
#define	CMD_VERIFY           ((char)'F') /* verify flash with arbitrary offset */
#define	CMD_DUMP             ((char)'D') /* dump rx buffer */
#define	CMD_SDRAM            ((char)'S') /* SDRAM read test */
#define	CMD_SDRAM_READ       ((char)'M') /* SDRAM test */
#define	CMD_SPI_PROGRAM      ((char)'L') /* SPI flash program */
#define	CMD_SPI_BOOT         ((char)'B') /* SPI flash Boot program offset 0x0000 */
#define	CMD_SPI_UBOOT        ((char)'U') /* SPI flash u-boot program offset 0x4000 */
#define	CMD_SPI_LOAD         ((char)'Z')
#define	CMD_SPI_VERIFY       ((char)'Y') /* SPI flash verify */
#define	CMD_NAND_PROGRAM     ((char)'N') /* NAND flash program */
#define	CMD_NAND_ECC_PROGRAM ((char)'C') /* NAND ECC flash program */
#define	CMD_NAND_JFFS2_PROGRAM ((char)'J') /* NAND JFFS2 flash program */
#define	CMD_NAND_VERIFY      ((char)'E') /* NAND flash program */
#define	CMD_NAND_RAW_PROGRAM ((char)'W') /* NAND flash raw program */
#define	CMD_NAND_RAW_VERIFY  ((char)'O') /* NAND flash raw verify */	
#define	CMD_NAND_DUMP        ((char)'P') /* NAND flash dump */
#define	CMD_NAND_ERASE       ((char)'A') /* NAND flash erase */	
#define	CMD_CONTINUE         ((char)'U')

#ifdef NANDWRITE
/* nand flash write utility - disable Data Flash Functions */
#undef CMD_SPI_PROGRAM
#undef CMD_SPI_BOOT
#undef CMD_SPI_UBOOT
#undef CMD_SPI_VERIFY
#endif 							/* NANDWRITE */

#ifndef NANDWRITE
#undef CMD_NAND_PROGRAM			
#undef CMD_NAND_VERIFY			
#undef CMD_NAND_RAW_PROGRAM
#undef CMD_NAND_ECC_PROGRAM
#undef CMD_NAND_JFFS2_PROGRAM
#undef CMD_NAND_RAW_VERIFY
#undef CMD_NAND_DUMP
#undef CMD_NAND_ERASE
#endif 							/* SPIBOOT */

#ifdef SPIBOOT
/* nand flash bootloader - disable Nand Flash Functions */
#undef CMD_NAND_PROGRAM
#undef CMD_NAND_VERIFY
#undef CMD_NAND_RAW_PROGRAM
#undef CMD_NAND_RAW_VERIFY
#undef CMD_NAND_DUMP
#undef CMD_NAND_ERASE
#endif 							/* SPIBOOT */

#ifndef SPIBOOT
    AT91F_DBGU_Printk("\nSend file by Xmodem:\n");

    /* receive file to SDRAM buffer */
    st = xmodemPipe.Read(&xmodemPipe, sdram_addr, sz, AT91_XmodemProtocol, (void *) sdram_addr);
    while (EndOfTransmission != 1);

    /* */
    AT91F_DBGU_Printk("\nTx Compl\n");

    pos = sz - 1;
    while(pos >= 0) {
	if( *(sdram_addr + pos) != 0xFF ) {
	    break;
	}
	pos--;
    }
    pos++;

    if(pos != 0) {
	AT91F_DBGU_Printv("Rx bytes ", pos, "\n");
    }

    /* */
    EndOfTransmission = 0;
    if(xmodemPipe.Reset != 0) {
        xmodemPipe.Reset( & xmodemPipe );
    }

/*#define NO_MENU */ 
#ifndef NO_MENU 
    {
	int im;
	static char * menu[] = {
#ifdef CMD_PROGRAM_0
	"[0]program flash (offset 0x00000)",
#endif 							/* CMD_PROGRAM_0 */
#ifdef CMD_PROGRAM_1
	"[1]write u-boot  (offset 0x10000)",
#endif 							/* CMD_PROGRAM_1 */
#ifdef CMD_PROGRAM_2
	"[2]write image   (offset 0x20000)",
#endif 							/* CMD_PROGRAM_2 */
#ifdef CMD_PROGRAM
	"pro[G]ram flash",
#endif							/* CMD_PROGRAM */
#ifdef CMD_READ
	"[R]ead flash",
#endif							/* CMD_READ */
#ifdef CMD_VERIFY_0
	"[V]erify flash",
#endif 							/* CMD_VERIFY_0 */
#ifdef CMD_VERIFY_1
	"ver[I]fy u-boot  (offset 0x10000)",
#endif 							/* CMD_VERIFY_1 */
#ifdef CMD_VERIFY
	"veri[F]y flash",
#endif							/* CMD_VERIFY */
#ifdef CMD_DUMP
	"[D]ump received buffer",
#endif							/* CMD_DUMP */
#ifdef CMD_SDRAM
	"[S]DRAM test",
#endif 							/* CMD_SDRAM */
#ifdef CMD_SDRAM_READ
	"SDRA[M] read test",
#endif							/* CMD_SPI_PROGRAM */
#ifdef CMD_SPI_PROGRAM
	"SPI f[L]ash program",
#endif							/* CMD_SPI_PROGRAM */
#ifdef CMD_SPI_BOOT
	"SPI flash [B]oot program",
#endif							/* CMD_SPI_BOOT */
#ifdef CMD_SPI_UBOOT
	"SPI flash [U]oot program",
#endif							/* CMD_SPI_UBOOT */
#ifdef CMD_SPI_VERIFY
	"SPI flash verif[Y]",
#endif							/* CMD_SPI_VERIFY */
#ifdef CMD_NAND_PROGRAM
	"[N]and program",
#endif 							/* CMD_NAND_PROGRAM */
#ifdef CMD_NAND_ECC_PROGRAM 
	"NAND E[C]C flash program",
#endif							/* CMD_NAND_ECC_PROGRAM */
#ifdef CMD_NAND_JFFS2_PROGRAM
	"NAND [J]FFS2 flash program",
#endif 							/* CMD_NAND_JFFS2_PROGRAM */
#ifdef CMD_NAND_VERIFY
	"Nand v[E]rify",
#endif 							/* CMD_NAND_VERIFY */
#ifdef CMD_NAND_RAW_PROGRAM
	"Nand ra[W] program",
#endif 							/* CMD_NAND_RAW_PROGRAM */
#ifdef CMD_NAND_RAW_VERIFY
	"Nand raw (data+[O]ob) verify",
#endif							/* CMD_NAND_RAW_VERIFY */
#ifdef CMD_NAND_DUMP
	"Nand dum[P]",
#endif							/* CMD_NAND_DUMP */
#ifdef CMD_NAND_ERASE
	"Nand er[A]se",
#endif							/* CMD_NAND_ERASE */

#ifdef CMD_CONTINUE
	"[C]ontinue",
#endif							/* CMD_CONTINUE */
	0};

	im = 0;
	while(menu[im] != 0) {
	    AT91F_DBGU_Printk(menu[im++]);
	    AT91F_DBGU_Printk("\n");
	}
    }
#endif 					/* NO_MENU */

    while(1) {
	it = 0;
	if(AT91F_DBGU_getch( & it )) {
	    if(it >= 'a' && it <= 'z') {
		it = 'A' + (it - 'a');
	    }
		
	    if(
#ifdef CMD_ERASE
	       (it == CMD_ERASE) ||		/* erase flash */
#endif 						/* CMD_ERASE */
#ifdef CMD_PROGRAM_0
	       (it == CMD_PROGRAM_0) ||		/* program flash from offset 0 */
#endif 						/* CMD_PROGRAM_0 */
#ifdef CMD_PROGRAM_1
	       (it == CMD_PROGRAM_1) ||		/* program flash from offset 0x10000 */
#endif 						/* CMD_PROGRAM_1 */
#ifdef CMD_PROGRAM_2
	       (it == CMD_PROGRAM_2) || 	/* program flash from offset 0x20000 */
#endif 						/* CMD_PROGRAM_2 */
#ifdef CMD_PROGRAM
	       (it == CMD_PROGRAM) || 
#endif 						/* CMD_PROGRAM */
#ifdef CMD_READ
	       (it == CMD_READ) || 
#endif 						/* CMD_READ */
#ifdef CMD_VERIFY_0
	       (it == CMD_VERIFY_0) ||		/* verify flash with offset 0 */ 	
#endif 						/* CMD_VERIFY_0 */
#ifdef CMD_VERIFY_1
	       (it == CMD_VERIFY_1) ||	 	/* verify flash with offset 0x10000 */
#endif 						/* CMD_VERIFY_1 */
#ifdef CMD_VERIFY
	       (it == CMD_VERIFY) ||	 	/* with abitrary offset */
#endif 						/* CMD_VERIFY */
#ifdef CMD_DUMP
	       (it == CMD_DUMP) ||	 	/* */
#endif 						/* CMD_DUMP */
#ifdef CMD_SDRAM
	       (it == CMD_SDRAM) || 		/* SDRAM test */
#endif 						/* CMD_SDRAM */
#ifdef CMD_SDRAM_READ
	       (it == CMD_SDRAM_READ) || 	/* SDRAM read test */
#endif 						/* CMD_SDRAM_READ */
#ifdef CMD_SPI_PROGRAM
	       (it == CMD_SPI_PROGRAM) || 	/* SPI flash program */
#endif  					/* CMD_SPI_PROGRAM */
#ifdef CMD_SPI_BOOT
	       (it == CMD_SPI_BOOT)   || 	/* SPI flash Boot program */
#endif						/* CMD_SPI_BOOT */
#ifdef CMD_SPI_UBOOT
	       (it == CMD_SPI_UBOOT)  ||
#endif
#ifdef CMD_SPI_VERIFY
	       (it == CMD_SPI_VERIFY) ||	/* SPI flash verify */
#endif 						/* CMD_SPI_VERIFY */
#ifdef CMD_NAND_PROGRAM
	       (it == CMD_NAND_PROGRAM) ||
#endif 						/* CMD_NAND_PROGRAM */
#ifdef CMD_NAND_ECC_PROGRAM
	       (it == CMD_NAND_ECC_PROGRAM) ||
#endif						/* CMD_NAND_ECC_PROGRAM */
#ifdef CMD_NAND_JFFS2_PROGRAM
	       (it == CMD_NAND_JFFS2_PROGRAM) ||
#endif						/* CMD_NAND_JFFS2_PROGRAM */
#ifdef CMD_NAND_VERIFY
	       (it == CMD_NAND_VERIFY) ||
#endif 						/* CMD_NAND_VERIFY */
#ifdef CMD_NAND_RAW_PROGRAM
	       (it == CMD_NAND_RAW_PROGRAM) ||
#endif						/* CMD_NAND_RAW_PROGRAM */
#ifdef CMD_NAND_RAW_VERIFY
	       (it == CMD_NAND_RAW_VERIFY) ||
#endif						/* CMD_NAND_RAW_VERIFY */
#ifdef CMD_NAND_ERASE
	       (it == CMD_NAND_ERASE) ||
#endif						/* CMD_NAND_ERASE */
#ifdef CMD_NAND_DUMP
	       (it == CMD_NAND_DUMP) ||
#endif						/* CMD_NAND_DUMP */

	       (it == CMD_CONTINUE)
	    ) {
		break;
	    } 
	}
    }


#ifdef CMD_SDRAM
    if(it == CMD_SDRAM) {
    /* SDRAM test */
	sdram_test(1); 
	continue;
    }
#endif 							/* CMD_SDRAM */

#else 				
    it = CMD_SPI_LOAD;				/* SPI flash program */
#endif						/* SPIBOOT */


#ifdef CMD_NAND_PROGRAM
    if(
#ifdef CMD_NAND_ERASE
	(it == CMD_NAND_ERASE) ||
#endif					/* CMD_NAND_ERASE */
#ifdef CMD_NAND_ECC_PROGRAM
	(it == CMD_NAND_ECC_PROGRAM) ||	/* Nand flash ECC program */	
#endif					/* CMD_NAND_ECC_PROGRAM */
#ifdef CMD_NAND_JFFS2_PROGRAM
	(it == CMD_NAND_JFFS2_PROGRAM) ||
#endif					/* CMD_NAND_JFFS2_PROGRAM */
#ifdef CMD_NAND_RAW_PROGRAM
	(it == CMD_NAND_RAW_PROGRAM) ||	/* Nand flash raw program */
#endif 					/* CMD_NAND_RAW_PROGRAM */
	(it == CMD_NAND_PROGRAM) 	/* NAND flash program */
    ) {
	int addr;			/* lin address */
	int row;			/* row (page) number */
	int fsize;			/* flash size */
	int bsize;			/* block size */
	int psize;			/* page size */
	int osize;			/* oob data size */
	int ppb;			/* pages per block */

	/* */
	fsize = (1 << nand_size);	/* flash size */
	bsize = (1 << nand_block);	/* block size */
	psize = (1 << nand_page);	/* page size */
	osize = (1 << nand_oob);	/* oob data size */

	/* pages per block */
	ppb = (1 << (nand_block - nand_page));

	/* */
        AT91F_DBGU_Printk("Nand erase\n");
	
	for(row = 0, addr = 0; addr < fsize; addr += bsize, row += ppb) {
	    /* */
	    AT91F_DBGU_Printx(addr);

	    if(it != CMD_NAND_ERASE) {
	    /* program, raw program, */
		unsigned char * oobbuf;
		oobbuf = ((unsigned char *) sdram_addr) + 0xF00000;
		if(nand_read((row+0), (-1), oobbuf, osize) != osize) { 
		    break;
		}
		if(nand_read((row+1), (-1), oobbuf + osize, osize) != osize) { 
		    break;
		}
	    
		if( *(oobbuf + 0) != 0xFF || *(oobbuf + osize) != 0xFF) {
	    	    AT91F_DBGU_Printk("Not valid block\n");
		    break;
		}
	    }

	    if(nand_erase(row) < 0) {
		break;
	    }
	}

	if(addr < fsize) {
	    AT91F_DBGU_Printx(row);
    	    AT91F_DBGU_Printk("Erase error\n");
	    continue;
	}
	
	if(it == CMD_NAND_ERASE) {
	    continue;
	}


        AT91F_DBGU_Printk("Nand ");
#ifdef CMD_NAND_ECC_PROGRAM
	if(it == CMD_NAND_ECC_PROGRAM) {	/* Nand flash program */
    	    AT91F_DBGU_Printk("ECC ");
	}
#endif 						/* CMD_NAND_ECC_PROGRAM */
#ifdef CMD_NAND_ECC_JFFS2_PROGRAM
	if(it == CMD_NAND_ECC_JFFS2_PROGRAM) {	/* Nand flash JFFS2 program */
    	    AT91F_DBGU_Printk("JFFS2 ");
	}
#endif 						/* CMD_NAND_ECC_PROGRAM */
#ifdef CMD_NAND_RAW_PROGRAM
	if(it == CMD_NAND_RAW_PROGRAM) {
    	    AT91F_DBGU_Printk("Raw ");
	}
#endif						/* CMD_NAND_RAW_PROGRAM */
        sz = pos;
	/* bytes */
        AT91F_DBGU_Printv("program", sz, "bytes\n");

	row = pos = 0;
	while(pos < sz) {
	    AT91F_DBGU_Printx(pos);
	    if(nand_write(row, 0,
	       ((unsigned char *) sdram_addr) + pos, psize) != psize) { 
		break;
	    }

#ifdef CMD_NAND_ECC_PROGRAM
	    if( (it == CMD_NAND_ECC_PROGRAM) ||	/* ECC program */
		(it == CMD_NAND_JFFS2_PROGRAM)	/* ECC JFFS2 program */
	    ) {
	    /* write oob */
		unsigned char * oobbuf;
		unsigned char ecc[3];
		int epos;
		int opos;
		oobbuf = ((unsigned char *) sdram_addr) + 0xF00000;
		/* clear page oob buffer */
		for(opos = 0; opos < 1024; ++opos) {
		    *(oobbuf + opos) = 0xFF;
		}

		epos = 0;
		opos = 0;
		while(epos < psize) {
		    int empty;
		    for(empty = 0; empty < 256; ++empty) {
			if(*(((unsigned char *) sdram_addr) + pos + epos + empty)
			 != 0xFF) { break; }
		    }
		    

		    if(empty < 256 && pos + epos + 256 <= sz) {
		    /* not empty buffer & full 256 bytes */
			ecc[0] = ecc[1] = ecc[2] = 0;
		    
			nand_calculate_ecc (
			    ((unsigned char *) sdram_addr) + pos + epos, ecc);

			if(it == CMD_NAND_ECC_PROGRAM) {
			    int ep;
			    for(ep = 0; ep < 3; ++ep) {
			        *(oobbuf + opos + ep) = ecc[ep];
			    }
			} else 
			if(it == CMD_NAND_JFFS2_PROGRAM) {
		    /* JFFS2 ecc bytes layout */
#define ALT_JFFS2_ECC_ORDER
#ifdef ALT_JFFS2_ECC_ORDER
			    *(oobbuf + opos + 5) = ecc[0];	/* 0 POS0 */
#else
			    *(oobbuf + opos + 0) = ecc[0];	/* 0 POS0 */
#endif							/* ALT_JFFS2_ECC_PRDER */
			    *(oobbuf + opos + 1) = ecc[1];	/* 1 POS1 */
			    *(oobbuf + opos + 2) = ecc[2];	/* 2 POS2 */
				/* 4 ECC valid */
			    *(oobbuf + opos + 4) = 0xF0;
			}		    
		    }

		    /* */
		    epos += 256;

		    /* */
		    for(empty = 0; empty < 256; ++empty) {
			if(*(((unsigned char *) sdram_addr) + pos + epos + empty)
			 != 0xFF) { break; }
		    }


		    if(empty < 256 && pos + epos + 256 <= sz) {
			ecc[0] = ecc[1] = ecc[2] = 0;

			nand_calculate_ecc (
			    ((unsigned char *) sdram_addr) + pos + epos, ecc);

			if(it == CMD_NAND_ECC_PROGRAM) {
			    int ep;
			    for(ep = 0; ep < 3; ++ep) {
				*(oobbuf + opos + 3 + ep) = ecc[ep];
			    }
			} else 
			if(it == CMD_NAND_JFFS2_PROGRAM) {
		    /* JFFS2 ecc bytes layout */
			    *(oobbuf + opos + 3) = ecc[0];	/* 3 POS3 */
			    *(oobbuf + opos + 6) = ecc[1];	/* 6 POS4 */
			    *(oobbuf + opos + 7) = ecc[2];	/* 7 POS5 */
			    *(oobbuf + opos + 4) &= 0x0F;	/* 4 ECC valid */
			}
		    }

		    /* */
		    epos += 256;
		    opos += 16;

		    /* */
		    if(nand_write(row, (-1), oobbuf, osize) != osize) { 
			break;
		    }
		}

#if 1
		if(row < 10) {
		    AT91F_DBGU_Printk("\nOOB:[\n");
		    for(opos = 0; opos < 64; ++opos) {
		        AT91F_DBGU_Printx(*(oobbuf + opos));
		    }
		    AT91F_DBGU_Printk("]\n");
		}
#endif
	    }
#endif 						/* CMD_NAND_ECC_PROGRAM */

	    pos += psize;

#ifdef CMD_NAND_RAW_PROGRAM
	    if(it == CMD_NAND_RAW_PROGRAM) {
	    /* write oob data */
		if(nand_write(row, (-1),
	    	    ((unsigned char *) sdram_addr) + pos, osize) != osize) { 
		    break;
		}
		pos += osize;
	    }
#endif							/* CMD_NAND_RAW_PROGRAM */

	    row++;			/* next page */
	}
	
	if(pos < sz) {
	    AT91F_DBGU_Printx(pos);
    	    AT91F_DBGU_Printk("Error\n");
	} else {
    	    AT91F_DBGU_Printk("Ok\n");
	}
	/* */
	continue;
    }
#endif 							/* CMD_NAND_PROGRAM */
#ifdef CMD_NAND_VERIFY
    if(
       (it == CMD_NAND_VERIFY) ||		/* NAND flash verify */
       (it == CMD_NAND_RAW_VERIFY)		/* NAND flash raw verify */
    ) {
	int row;			/* row (page) number */
	int psize;			/* page size */
	int osize;			/* oob data size */

        AT91F_DBGU_Printk("Nand verify\n");

        sz = pos;			
	psize = (1 << nand_page);	/* page size */
	osize = (1 << nand_oob);	/* oob data size */

	row = 0;			/* page number */
	pos = 0;			/* sdram buffer position */
	while(pos < sz) {
	    int len;
	    int ps;

	    /* */    
	    AT91F_DBGU_Printx(pos);

	    /* clear buffer */
	    for(ps = 0; ps < (psize + osize); ++ps) {
	       *(((unsigned char *) sdram_addr) + 0xF00000 + ps) = 0;
	    }

	    /* read usual data */
	    if(nand_read(row, 0,
	       ((unsigned char *) sdram_addr) + 0xF00000, psize) != psize) { 
		{ break; }
	    }

	    if(it == CMD_NAND_RAW_VERIFY) {
	    /* read oob data */
		if(nand_read(row, (-1),
	    	    ((unsigned char *) sdram_addr) + 0xF00000 + psize, osize) != osize) { 
		    { break; }
		}
	    }

	    len = sz - pos;
	    if(it == CMD_NAND_RAW_VERIFY) {
		if(len > (psize+osize)) {
		    len = (psize+osize);	/* page size with oob data */
		}
	    } else {
		if(len > psize) {
		    len = psize;		/* page size without oob data */
		}
	    }

	    for(ps = 0; ps < len; ++ps) {
		unsigned char chw;
		unsigned char chr;
		/* writed character */
		chw = *(((unsigned char *) sdram_addr) + (pos + ps));
		/* readed character */
		chr = *(((unsigned char *) sdram_addr) + (0xF00000 + ps));
		if(chw != chr) {
		    AT91F_DBGU_Printk("Error ");
		    AT91F_DBGU_Printw("offset ", pos + ps, "\n");
		    AT91F_DBGU_Printv("byte writed ", chw, "\n");
		    AT91F_DBGU_Printv("byte readed ", chr, "\n");
//		    break;
		}
	    }

	    pos += psize;
	    if(it == CMD_NAND_RAW_VERIFY) {
		pos += osize;
	    }
	    row++;
	} 
	
	if(pos >= sz) {
    	    AT91F_DBGU_Printk("Ok\n");
	}

	continue;
    }
#endif 							/* CMD_NAND_VERIFY */

#ifdef CMD_NAND_DUMP
    if(it == CMD_NAND_DUMP) {
	unsigned long page;		/* page */	
	int osize;			/* oob size */
	int psize;			/* page size */
	int off;			/* page offset */
	char * buf;
	char more;
	
	AT91F_DBGU_Printk("page number_");
	AT91F_DBGU_getx(& page);			/* read address from console */
	AT91F_DBGU_Printk("\n");

	buf = ((unsigned char *) sdram_addr) + 0xF00000;
		
	psize = (1 << nand_page);	/* page size */
	osize = (1 << nand_oob);	/* oob data size */

	more = 1;
	while(more) {	
	    AT91F_DBGU_Printw("Page",page,"\n");
	    off = 0;	
	    while(off < psize && more) {

		if(nand_read(page, off, buf, 256) != 256) { 
		    more = 0;
	        } else {
		    if(AT91_dump(buf, 256) != 1) {
			more = 0;
			break;
		    }
		}
		off += 256;
	    }

	    if(more == 1) {
		if(nand_read(page, (-1), buf, osize) != osize) { 
		    more = 0;
		} else {
		    AT91F_DBGU_Printk("OOB\n");
		    if(AT91_dump(buf, osize) != 1) {
			more = 0;
		    }
		}
	    }
	    page++;
	}

	continue;
    }
#endif							/* CMD_NAND_DUMP */


#ifdef CMD_SDRAM_READ
    if(it == CMD_SDRAM_READ) {
    /* SDRAM read test */
	unsigned long sdram_addr;				/* SDRAM address */
	unsigned long value;					/* readed value */

	sdram_addr = 0;
	AT91F_DBGU_Printk("SDRAM test (infinite loop)\n");
	AT91F_DBGU_Printk("SDRAM offset_");
	AT91F_DBGU_getx(& sdram_addr);			/* read address from console */
	AT91F_DBGU_Printk("\n");
	if(sdram_addr >= sdram_size) {
	    AT91F_DBGU_Printk("Incorrect SDRAM offset\n");
	    continue;
	} 
	sdram_addr |= 0x20000000;
	AT91F_DBGU_Printw("SDRAM offset ", sdram_addr, "\n");

	while(1) {								/* infinite loop  */
	    value =  *((volatile unsigned long *) sdram_addr);
	}
	continue;
    }
#endif 							/* CMD_SDRAM_READ */

    if(it == CMD_CONTINUE) {					// receive next file
	continue;
    } 

#if (defined(CMD_READ) || defined(CMD_DUMP))
    if(
#ifdef CMD_READ
        (it == CMD_READ)
#ifdef CMD_DUMP
         || 
#endif 						/* CMD_DUMP */
#endif 						/* CMD_READ */
#ifdef CMD_DUMP
        (it == CMD_DUMP)
#endif 						/* CMD_DUMP */
    ) {
    /* read flash OR dump flash */
	unsigned char * addr;		// address
	int col, row;			// column, address 
	int off;			// flash or buffer offset

#ifdef CMD_READ
    	if(it == CMD_READ) {
	/* read flash */
    	    addr = (unsigned char *) FLASH_ADDR;
	}
#endif 						/* CMD_READ */

#ifdef CMD_DUMP
    	if(it == CMD_DUMP) {
	/* dump rx buffer */
	    addr = (unsigned char *) SDRAM_ADDR;
	}
#endif  					/* CMD_DUMP */

	while(1) {
	    if(AT91_dump(addr,256) == 0) {
		break;
	    }
	}
	continue;
    }
#endif 					/* CMD_READ || CMD_DUMP */
		 

#if (defined(CMD_VERIFY_0) || defined(CMD_VERIFY_1) || defined(CMD_VERIFY))
    if(
       (it == CMD_VERIFY_0) || 
       (it == CMD_VERIFY_1) || 
       (it == CMD_VERIFY)
    ) {

/* verify flash command */
        unsigned char * flash_addr = (unsigned char *) FLASH_ADDR;
	int sz = pos; 
	/* */
	AT91F_DBGU_Printk("Verify flash\n");
	if(it == CMD_VERIFY_1) {
	/* verify flash from offset 0x10000 */
	    AT91F_DBGU_Printk("Offset 0x10000\n");
	    flash_addr += 0x10000;
	} else
	if(it == CMD_VERIFY) {
	    unsigned long off = 0;
	    AT91F_DBGU_Printk("Input start offset (0x0000)_");
	    AT91F_DBGU_getx(& off);
	    flash_addr += off;
	    AT91F_DBGU_Printk("\n");
	}	

	for(pos = 0; pos < sz; ++pos) {
	    if( *(sdram_addr + pos) != *(flash_addr + pos) ) {
	        AT91F_DBGU_Printv("Error offset [", pos, "]\n");
	        break;
	    }
    	}

	if(pos >= sz) {
	    AT91F_DBGU_Printk("Ok\n");
	} 
	/* */
	{ continue; }
    }
#endif	 						/* CMD_VERIFY */

#ifdef CMD_SPI_PROGRAM
    if(
	(it == CMD_SPI_PROGRAM)	||	/* raw SPI flash program */
        (it == CMD_SPI_BOOT) 	||	/* SPI flash Boot program */
        (it == CMD_SPI_UBOOT)		/* SPI flash u-boot program */
    ) { 
/* SPI flash program */	
        unsigned char * buf;		// data buffer 
	unsigned int w_crc;		// writed data CRC
	unsigned int r_crc;		// readed data CRC
	int status;			// flash access status
	int psz;			// page size
	int pnm;			// page number 

    	AT91F_DBGU_Printk("\nSPI flash ");
        if(it == CMD_SPI_BOOT) {
    	    AT91F_DBGU_Printk("Boot ");
	} else 
        if(it == CMD_SPI_UBOOT) {
    	    AT91F_DBGU_Printk("u-boot ");
	}
    	AT91F_DBGU_Printk("program\n");

	sz = pos;	    				// bytes to write  
	buf = (unsigned char *) sdram_addr;

        if(it == CMD_SPI_BOOT) {	/* SPI flash Boot program */
	    unsigned long sixvec;

	    if(sz > SPI_BOOT_SIZE) {			/* bits 0-7 */
		sz = SPI_BOOT_SIZE;
	    }
	    sixvec = ((sz + 511) >> 9);	/* block's load to internal SRAM  */

	    /* bits 13-16 number of pages 2**n */
	    if(DeviceAT45DB.pages_number == 8192) {	/* 2**13 */
		sixvec |= (13 << 13);
	    } else {
		sixvec |= (13 << 13);			/* 2**13 */	
	    }

	    sixvec |= (DeviceAT45DB.pages_size << 17);
	
	    /* write six vector */
	    *(((unsigned long *) buf) + 5) = sixvec;
	}

	/* calculate CRC  */
	w_crc = 0xFFFFFFFF;
	pAT91->CRC32( (unsigned char *) sdram_addr, sz, &w_crc);

	AT91F_DBGU_Printw("CRC32=", w_crc, "\n");

	pnm = 0;
        if(it == CMD_SPI_UBOOT) {
	    for(pos = 0; pos < SPI_BOOT_SIZE; pos += DeviceAT45DB.pages_size) {
		pnm++;
	    }
	}

	psz = DeviceAT45DB.pages_size;
	for(pos = 0; pos < sz; buf += psz, pos += psz, ++pnm) {

	    /* erase page  */
	    status = svcDataFlash.PageErase( &svcDataFlash, pnm);
	    AT91F_DataFlashWaitReady(AT91C_DATAFLASH_TIMEOUT);

	    if(status == 0x1) {
	    } else {
	    /* print page number */
		AT91F_DBGU_Printv("\nPage [", pnm, "] Erase error.");
		AT91F_DBGU_Printw("\nStatus=", status, "\n");
		break;	
	    }
	
	    /* page program  */
	    status = svcDataFlash.PagePgmBuf( &svcDataFlash,
	     DB_PAGE_PGM_BUF1, buf, pnm * psz, psz);

	    /* Wait DataFlash Ready */
	    AT91F_DataFlashWaitReady(AT91C_DATAFLASH_TIMEOUT);

	    if(status == 0x1) {
		AT91F_DBGU_Printw(0,pos,0);
	    } else { 
	        AT91F_DBGU_Printw("\nWrite Failed. Status=", status, "");
	        AT91F_DBGU_Printw(" Offset=", pos, "\n");
	        break;	
	    }
	}

	if(pos < sz) {
	/* SPI flash program error */
	    continue;
	}

    	AT91F_DBGU_Printk("SPI flash verify\n");

	/* clear input buffer */
	buf = ((unsigned char *) sdram_addr) + 0xF00000;
	for(pos = 0; pos < sz; ++pos) {
	    *(buf + pos) = 0;
	}

	pnm = 0;
        if(it == CMD_SPI_UBOOT) {
	    for(pos = 0; pos < SPI_BOOT_SIZE; pos += DeviceAT45DB.pages_size) {
		pnm++;
	    }
	}

	for(pos = 0; pos < sz; buf += psz, pos += psz, pnm++) {
	
	    status = svcDataFlash.PageRead( &svcDataFlash,
 	        pnm * psz, buf, psz);

	    /* Wait DataFlash Ready */
	    AT91F_DataFlashWaitReady(AT91C_DATAFLASH_TIMEOUT);

	    if(status == 0x1) {
	    } else { 
	        AT91F_DBGU_Printw("\nRead Failed. Status=", status, "");
	        AT91F_DBGU_Printw(" Offset=", pos, "\n");
	        break;	
	    }
	}

	if(pos < sz) {
	/* SPI flash read error */
	    continue;
	}

	r_crc = 0xFFFFFFFF;
	pAT91->CRC32( ((unsigned char *) sdram_addr) + 0xF00000, sz, &r_crc);
		
    	AT91F_DBGU_Printw("\nCRC32=", r_crc, "\n");

	if(w_crc == r_crc) {
    	    AT91F_DBGU_Printk("Ok\n");
	} else {
    	    AT91F_DBGU_Printk("Readed CRC not equal Writed CRC.\n");
	}

	{ continue; }
    }
#endif 												/* CMD_SPI_PROGRAM */

#ifdef CMD_SPI_VERIFY
    if(
        (it == CMD_SPI_VERIFY) ||
        (it == CMD_SPI_LOAD)
    ) { 
        unsigned char * buf;			// SDRAM buffer 
	unsigned int psz;				// page size 
	int i, status;
	int pnm;

        if(it == CMD_SPI_VERIFY) {
    	    AT91F_DBGU_Printk("\nSPI flash verify\n");
	} else {
    	    AT91F_DBGU_Printk("Load u-boot from SPI flash to SDRAM\n");
	}

//	
	sz = pos;						// save size 
	pnm = 0;
	buf = (unsigned char *) sdram_addr;

        if(it == CMD_SPI_LOAD) {
	    buf += 0xF00000;

	    pnm = 0;
	    for(pos = 0; pos < SPI_BOOT_SIZE; pos += DeviceAT45DB.pages_size) {
		++pnm;
	    }
	    sz = 128 * 1024;
	}
/* */
	psz = DeviceAT45DB.pages_size; 
/* */
	for(pos = 0; pos < sz; pos += psz,pnm++) {
	
	    status = svcDataFlash.PageRead( &svcDataFlash,
		pnm*psz, Buffer_Dataflash, psz);

	    /* Wait DataFlash Ready */
	    AT91F_DataFlashWaitReady(AT91C_DATAFLASH_TIMEOUT);

	    if(status == 0x1) {
	        for(i = 0; i < psz && pos + i < sz; ++i) {
		    unsigned char ch;
		    ch = *((unsigned char *) Buffer_Dataflash + i);
		    if(it == CMD_SPI_LOAD) {
			*(buf + pos + i) = ch;
		    } else {
			if( *(buf + pos + i) !=  ch) {
		    	    break;
			}
		    }
		}

		if(i < psz && pos + i < sz) {
		    AT91F_DBGU_Printw("\nCompare failure. Offset=", pos + i, "");
		    AT91F_DBGU_Printw("\n Size", sz, "\n");
		    break;
		}
	    } else { 
	        AT91F_DBGU_Printw("\nRead Failed. Status=", status, "");
	        AT91F_DBGU_Printw(" Offset=", pos, "\n");
	        break;
	    }
        }

	if(it == CMD_SPI_LOAD) {
/* */
	    AT91F_AIC_DisableIt (AT91C_BASE_AIC, AT91C_ID_SYS);
	    /* close all peripheral clocks */
	    AT91C_BASE_PMC->PMC_PCDR = 0xFFFFFFFC;
	    /* Disable core interrupts and set supervisor mode */
	    __asm__ ("msr CPSR_c, #0xDF");
	    //* Clear all the interrupts */
	    *AT91C_AIC_ICCR = 0xffffffff;
	    /* read the AIC_IVR and AIC_FVR */
	    i = *AT91C_AIC_IVR;
	    i = *AT91C_AIC_FVR;
	    /* write the end of interrupt control register */
	    *AT91C_AIC_EOICR = 0;
	    AT91F_DBGU_Printk("Run uboot\n");
	    {
		unsigned char * dst = 
    		    (unsigned char *) ( 0x20000000 + 0xF00000);
		asm("mov pc,%0" : : "r" (dst));
	    }
	} else {
	    if(pos >= sz) {
		AT91F_DBGU_Printk("Ok\n");
	    }
	}    	

	{ continue; }
    }
#endif 													/* CMD_SPI_VERIFY */

#if (defined(CMD_ERASE) || defined(CMD_PROGRAM_0))
    if(it == CMD_ERASE || it == CMD_PROGRAM_0) {
/* erase flash command */
#ifdef FLASH_16
	volatile unsigned short * flash_addr = (volatile unsigned short *) FLASH_ADDR;
	unsigned short chr;				/* */

	AT91F_DBGU_Printk("Erase flash\n");

	*(flash_addr + 0x555) = 0xAA;	/* 1 write command */
	*(flash_addr + 0x2AA) = 0x55;	/* 2 */
	*(flash_addr + 0x555) = 0x80;	/* 3 */
	*(flash_addr + 0x555) = 0xAA;	/* 4 */
	*(flash_addr + 0x2AA) = 0x55;	/* 5 */
	*(flash_addr + 0x555) = 0x10;	/* 6 */

	do {
	    chr = *(flash_addr + 0);
	    chr &= 0x80;
	} while(chr == 0);

        *(flash_addr + 0x0)   = 0xF0;	/* write reset command */
	AT91F_DBGU_Printk("Ok\n");
#else 
	/* erase in 8 bit mode */

	volatile unsigned char * flash_addr = (volatile unsigned char *) FLASH_ADDR;
	unsigned char chr;			/* */

	*(flash_addr + 0xAAA) = 0xAA;	/* 1 write command */
	*(flash_addr + 0x555) = 0x55;	/* 2 */
	*(flash_addr + 0xAAA) = 0x80;	/* 3 */
	*(flash_addr + 0xAAA) = 0xAA;	/* 4 */
	*(flash_addr + 0x555) = 0x55;	/* 5 */
	*(flash_addr + 0xAAA) = 0x10;	/* 6 */

	do {
 	    chr = *(flash_addr + 0);
  	    chr &= 0x80;
  	} while(chr == 0);

	*(flash_addr + 0x0)   = 0xF0;	/* write reset command */
	AT91F_DBGU_Printk("Ok\n");
#endif  											/* FLASH_16 */
	/* */
	if(it == CMD_ERASE) {	
	    continue;
	}
    }
#endif 							/* CMD_ERASE */

#if (defined(CMD_PROGRAM_0) || defined(CMD_PROGRAM_1) || defined(CMD_PROGRAM_2))
    if(it == CMD_PROGRAM_0 || it == CMD_PROGRAM_1 || it == CMD_PROGRAM_2) {
        AT91F_DBGU_Printk("Write flash \n");
        AT91F_DBGU_Printk("Offset ");
        switch(it) {
    	    case CMD_PROGRAM_0 : AT91F_DBGU_Printk( "0x0000"); break;
	    case CMD_PROGRAM_1 : AT91F_DBGU_Printk("0x10000"); break;
	    case CMD_PROGRAM_2 : AT91F_DBGU_Printk("0x20000"); break;
	}
	AT91F_DBGU_Printk("\n");

#ifdef FLASH_16
	{
   	    volatile unsigned short * flash_addr = (volatile unsigned short *) FLASH_ADDR;
	    int sz;

	    sz = 0;
	    if(it == CMD_PROGRAM) {
	        unsigned long off = 0;
	        AT91F_DBGU_Printk("Program flash. Input start offset (0x0000)_");
	        AT91F_DBGU_getx(& off);
	        sz = off;
	        AT91F_DBGU_Printk("\n");
	    } else
	    if(it == CMD_PROGRAM_1) {				
	        sz = 0x10000;					/* u-boot offset */
	    } else 
	    if(it == CMD_PROGRAM_2) {			/* image offset */
	        sz = 0x20000;
	    }

	    /* */
	    flash_addr = (unsigned short *) (FLASH_ADDR + sz);

	    sz = (pos + 1) / 2;
	
	    for(pos = 0; pos < sz; ++pos) {
		unsigned short chr;
	        unsigned short ch;

		ch = *( ((unsigned short *) sdram_addr) + pos );

	    	/* */
   		*(flash_addr + 0x555) = 0xAA;
 		*(flash_addr + 0x2AA) = 0x55;
		*(flash_addr + 0x555) = 0xA0;
		*(flash_addr + pos) = ch;

		do {
		    chr = *(flash_addr + pos);
		    if( (chr & 0x80) == (ch & 0x80) ) {
		        break;
		    } 
   		} while( (chr & 0x20) == 0 );


	    	chr = *(flash_addr + pos);

		/* */
		if( (chr & 0x80) != (ch & 0x80) ) {
		    AT91F_DBGU_Printk("Fail");
 	   	    break;
		}
	    }

	    /* write reset command */
    	    *(flash_addr + 0x0) = 0xF0;	
	
	    if(pos >= sz) {
	        AT91F_DBGU_Printk("Ok\n");
	    }

	}
#else 
	{
	    volatile unsigned char * flash_addr = (volatile unsigned char *) FLASH_ADDR;
	    int sz;

	    if(it == CMD_PROGRAM_1) {				
    	        flash_addr += 0x10000;	
	    } else 
	    if(it == CMD_PROGRAM_2) {	
	        flash_addr += 0x20000;
	    }

	    sz = pos;
	    for(pos = 0; pos < sz; ++pos) {
	        unsigned char chr;
	        unsigned char ch;

	        ch = *(sdram_addr + pos);

	        /* */
   	        *(flash_addr + 0xAAA) = 0xAA;
 	        *(flash_addr + 0x555) = 0x55;
	        *(flash_addr + 0xAAA) = 0xA0;
	        *(flash_addr + pos) = ch;
    	        do {
		    chr = *(flash_addr + pos);
		    if( (chr & 0x80) == (ch & 0x80) ) {
		        break;
		    } 
   		} while( (chr & 0x20) == 0 );

    		chr = *(flash_addr + pos);
    		/* */
	
		if( (chr & 0x80) != (ch & 0x80) ) {
		    AT91F_DBGU_Printk("Fail");
 	   	    break;
		}
	    }

	    /* write reset command */
    	    *(flash_addr + 0x0)   = 0xF0;

	    if(pos >= sz) {
	        AT91F_DBGU_Printk("Ok\n");
	    }

	}
#endif 							/* FLASH_16 */
    }
#endif 							/* CMD_PROGRAM_0 */
 }
 return 0;	
}

